# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/ml.llm.utils.ipynb.

# %% auto 0
__all__ = ['LLAMA2_CHAT_PROMPT_TEMPLATE']

# %% ../../../nbs/ml.llm.utils.ipynb 3
LLAMA2_CHAT_PROMPT_TEMPLATE = """<s>[INST] <<SYS>>
{system_prompt}
<</SYS>>

{user_message} [/INST]"""
